// Copyright (C) 2006 Erik Dahlberg
// 
// This file is part of LSystem3D.
// 
// LSystem3D is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// LSystem3D is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with LSystem3D; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA




#include <sstream>
#include <string>

#include "fx.h"

#include "gui.h"
#include "lindenmayersystem.h"
#include "lsystemparameters.h"
#include "renderingsurface.h"
#include "rule.h"
#include "ruleset.h"

using namespace std;



// message map
FXDEFMAP(GUI) GUIMap[] =
{
    FXMAPFUNC(SEL_COMMAND, GUI::ID_GENERATE, GUI::onGenerateButtonPressed),
    FXMAPFUNC(SEL_COMMAND, GUI::ID_HELP,     GUI::onHelpButtonPressed)
};

// macro to set up class implementation
FXIMPLEMENT(GUI, FXMainWindow, GUIMap, ARRAYNUMBER(GUIMap))



/**
 * Constructor
 * @param application the FOX application object
 * @param renderingSurface the rendering surface
 * @param lsystem the Lindenmayer-system
 */
GUI::GUI(FXApp *application, RenderingSurface *renderingSurface, LindenmayerSystem *lsystem) : FXMainWindow(application, "LSystem3D" , NULL, NULL, DECOR_ALL, 920, 100, 0, 0)
{
    _lsystem = lsystem;
    _renderingSurface = renderingSurface;
    
    
    
    // create the widgets
    
    // the main frame
    FXVerticalFrame *mainFrame = new FXVerticalFrame(this, LAYOUT_FILL_X);
    
    // axiom text field
    FXHorizontalFrame *axiomFrame = new FXHorizontalFrame(mainFrame, LAYOUT_FILL_X);
    new FXLabel(axiomFrame, "Axiom:");
    _axiomTextField = new FXTextField(axiomFrame, 10, NULL, 0, LAYOUT_FILL_X);
    
    // rules multiline text field
    FXHorizontalFrame *rulesFrame = new FXHorizontalFrame(mainFrame, LAYOUT_FILL_X);
    new FXLabel(rulesFrame, "Rules:");
    _rulesText = new FXText(rulesFrame, NULL, 0, LAYOUT_FILL_X);
    _rulesText->setVisibleColumns(30);
    _rulesText->setVisibleRows(10);
    
    // angle spinner
    FXHorizontalFrame *angleFrame = new FXHorizontalFrame(mainFrame);
    new FXLabel(angleFrame, "Angle:");
    _angleRealSpinner = new FXRealSpinner(angleFrame, 5);
    
    // depth spinner
    FXHorizontalFrame *depthFrame = new FXHorizontalFrame(mainFrame);
    new FXLabel(depthFrame, "Depth:");
    _depthSpinner = new FXSpinner(depthFrame, 5);
    
    // diameter spinner
    FXHorizontalFrame *diameterFrame = new FXHorizontalFrame(mainFrame);
    new FXLabel(diameterFrame, "Diameter:");
    _diameterRealSpinner = new FXRealSpinner(diameterFrame, 5);
    _diameterRealSpinner->setValue(0.2);
    _diameterRealSpinner->setIncrement(0.1);
    
    // diameter factor spinner
    new FXLabel(diameterFrame, "Factor:");
    _diameterFactorRealSpinner = new FXRealSpinner(diameterFrame, 5);
    _diameterFactorRealSpinner->setValue(0.9);
    _diameterFactorRealSpinner->setIncrement(0.01);
    
    // separator
    new FXHorizontalSeparator(mainFrame, SEPARATOR_RIDGE | LAYOUT_FILL_X);

    // generate button
    new FXButton(mainFrame, "&Generate", NULL, this, GUI::ID_GENERATE);
    
    // help button
    new FXButton(mainFrame, "&Help", NULL, this, GUI::ID_HELP);
    
    // help dialog
    stringstream helpText;
    helpText << "Mouse:" << endl
             << "Left button: Move model in z-plane" << endl
             << "Middle button: Zoom in/out" << endl
             << "Right button: Rotate around y-axis" << endl
             << endl
             << "Rules:" << endl
             << "F : Walk forward, leaving a trace" << endl
             << "A-Z : Replacement rule" << endl
             << "A(0.33) : Probability factor 0.33 for rule \"A\"" << endl
             << "= : Rule assignment" << endl
             << "+ : Turn left" << endl
             << "- : Turn right" << endl
             << "&& : Pitch down" << endl
             << "^ : Pitch up" << endl
             << "\\ : Roll left" << endl
             << "/ : Roll right" << endl
             << "| : Turn around 180 degrees" << endl
             << "[ : Save state to stack" << endl
             << "] : Load state from stack" << endl
             << "{ : Create a planar surface" << endl
             << "} : Close a planar surface" << endl
             << "f : One vertex in a planar surface, specified CCW" << endl
             << "! : Decrement diameter of segment" << endl
             << "\' : Increment current index to color table" << endl
             << ", : Decrement current index to color table" << endl
             << endl
             << "Example:" << endl
             << "F(0.33)=[+F/L]/F[-F&&L]F!" << endl
             << "F(0.33)=F[+F+L]&&F!" << endl
             << "F(0.34)=F[-F-L]/F!" << endl
             << "L=,{-f++f-|-f++f-}'";
    
    _helpMessageBox = new FXMessageBox(getApp(), "Help", helpText.str().c_str(), NULL, MBOX_OK);
    
    
    
    // load L-system parameters from file and sync with GUI
    
    _lsystemParameters.load(_lsystem, "lsystem.xml");
    
    // axiom
    _axiomTextField->setText(_lsystem->getAxiom().getContent().c_str());
    
    // rules
    rulemap rules = _lsystem->getRules().getRules();
    string rulesString;    
    for (rulemap::iterator it = rules.begin(); it != rules.end(); ++it)
    {
        rulesString += it->second.toString();
        rulesString += '\n';
    }
    _rulesText->setText(rulesString.c_str(), rulesString.size());
    
    // angle
    _angleRealSpinner->setValue(_lsystem->getAngle());
    
    // depth
    _depthSpinner->setValue(_lsystem->getDepth());
}



/**
 * Create and initialize the window
 */
void GUI::create()
{
    // create the window and make it appear
    FXMainWindow::create();
    show();
}



/**
 * Called by the system when the "Generate" button is pressed
 * @param sender the sender object
 * @param selector message type and id
 * @param data event related data
 * @return 
 */
long GUI::onGenerateButtonPressed(FXObject *sender, FXSelector selector, void *data)
{
    // clear L-system
    _lsystem->clear();
    
    
    // get user input form widgets
    
    // axiom
    Rule axiom("axiom", _axiomTextField->getText().text(), 1.0);
    _lsystem->setAxiom(axiom);
    
    // rules
    string ruleSet = _rulesText->getText().text();
    string currentRule;
    for (int i = 0; i < ruleSet.size(); i++)
    {
        if (ruleSet[i] != '\n')
        {
            // rule component
            currentRule += ruleSet[i];
            
            if (i == ruleSet.size() - 1)
            {
                // last char in whole rule string
                _lsystem->setRule(Rule(currentRule));
                currentRule.clear();
            }
        }
        else
        {
            // last char in current rule
            _lsystem->setRule(Rule(currentRule));
            currentRule.clear();
        }
    }

    // angle
    _lsystem->setAngle(_angleRealSpinner->getValue());
    
    // depth
    _lsystem->setDepth(_depthSpinner->getValue());
    
    // diameter
    _lsystem->setDiameter(_diameterRealSpinner->getValue());
    
    // diameter factor
    _lsystem->setDiameterFactor(_diameterFactorRealSpinner->getValue());
    
    
    // generate and render L-system to screen
    _lsystem->generate();
    _renderingSurface->draw();
}



/**
 * Called by the system when the "Help" button is pressed
 * @param sender the sender object
 * @param selector message type and id
 * @param data event related data
 * @return 
 */
long GUI::onHelpButtonPressed(FXObject *sender, FXSelector selector, void *data)
{
    _helpMessageBox->show(PLACEMENT_OWNER);
}
